<?php

namespace NoccyLabs\TinyApp;

use ReflectionFunction;

class TinyApp
{
    protected static $instance;

    protected $name;

    protected $info;

    protected $opts = [];

    protected $blocks = [];

    public static function getInstance()
    {
        if (!self::$instance) self::$instance = new self();
        return self::$instance;
    }

    protected function __construct()
    {}

    public function setName($name)
    {
        $this->name = $name;
    }

    public function setInfo($info)
    {
        $this->info = $info;
    }

    public function setVersion($version)
    {
        $this->version = $version;
    }

    public function addOptions(array $opts)
    {
        $this->opts = array_merge($this->opts,$opts);
    }

    public function addBlock($block)
    {
        $this->blocks[] = $block;
    }

    public function run(callable $entry, $flags=0)
    {
        $parser = new ArgumentParser();
        $parser->setOptions($this->opts);
        $parsed = $parser->getParsed();

        if ($flags & TINYAPP_HAS_HELP) {
            if (array_key_exists('help',$parsed) && ($parsed['help'])) {
                $this->printUsage();
                exit(0);
            }
        }

        $args = $this->getEntryArgs($entry, $parsed);

        try {
            $ret = call_user_func($entry, ...$args);
        } catch (\Exception $e) {
            if (!$flags & TINYAPP_EXCEPTION_GUARD)
                throw $e;
            $handler = new ExceptionHandler();
            $handler->printException($e);
        } catch (\Throwable $e) {
            if (!$flags & TINYAPP_EXCEPTION_GUARD)
                throw $e;
            $handler = new ExceptionHandler();
            $handler->printException($e);
        }
    }

    public function printUsage()
    {
        
        printf("\e[96m%s\e[36m%s\e[90m - \e[94m%s\e[0m\n", $this->name, $this->version?" {$this->version}":"", $this->info);
        printf("\n");
        if (count($this->blocks) > 0) {
            $this->printBlock($this->blocks[0]);
        }

        if (count($this->opts)>0) {
            foreach ($this->opts as $opt=>$info) {
                $this->printOption($opt, $info);
            }
            printf("\n");
        }
        foreach (array_slice($this->blocks,1) as $block) {
            $this->printBlock($block);
        }
    }

    private function printOption($option, $info)
    {
        if (strpos($option,":")!==false) {
            list($option, $arg) = explode(":", $option, 2);
        } else {
            $arg = null;
        }
        if (strpos($option,",")!==false) {
            list($option, $short) = explode(",", $option, 2);
        } else {
            $short = null;
        }
        $out = null; $len = 0;
        $out.= "--\e[1m{$option}\e[21m";
        $len+= 2 + strlen($option);
        if ($short) {
            $out.= ",-\e[1m{$short}\e[21m";
            $len+= 2 + strlen($short);
        }
        if ($arg) {
            $out.= " \e[4m{$arg}\e[24m";
        $len+= 1 + strlen($arg);
        }
        if ($len < 20) {
            $out.= str_repeat(" ",20-$len);
        }
        echo "  \e[32m{$out} \e[92m{$info}\e[0m\n";
    }

    private function printBlock($block)
    {
        if (!is_array($block)) {
            $blocks = (array)$block;
        } else {
            $blocks = $block;
        }
        foreach ($blocks as $block) {
            if (substr($block,0,2)=="//") {
                $style="\e[3m";
                $block = substr($block,2);
            } elseif (substr($block,0,2)=="**") {
                $style="\e[1m";
                $block = substr($block,2);
            } elseif (substr($block,0,2)=="!!") {
                $style="\e[31;1m";
                $block = substr($block,2);
            } else {
                $style="";
            }
            $block = wordwrap($block);
            echo $style.$block."\e[0m\n\n";
        }
    }

    private function getEntryArgs(callable $entry, array $opt_parsed)
    {
        // Get reflection of entry function to resolve option mapping
        $reflect = new ReflectionFunction($entry);
        $params = $reflect->getParameters();
        $call_args = [];
        foreach ($params as $param) {
            $name = $param->getName();
            $fname = str_replace("_","-",$name);
            //printf("name=%s, fixed=%s, opts=%s\n", $name, str_replace("_","-",$name), json_encode($opt_parsed));
            if (array_key_exists($fname, $opt_parsed)) {
                $arg = $opt_parsed[$fname];
                if ($param->isArray() && (!is_array($arg))) {
                    $arg = [ $arg ];
                } elseif ((!$param->isArray()) && is_array($arg)) {
                    fprintf(STDERR, "Error: --%s can only be specified once\n", $name);
                    exit(1);
                }
                $call_args[] = $arg;
            } elseif ($name == "_") {
                $call_args[] = $opt_parsed['_params'];
            } elseif ($name == "__") {
                $call_args[] = $opt_parsed;
            } else {
                $call_args[] = $param->isOptional()?$param->getDefaultValue():null;
            }
        }
        return $call_args;

    }

}

